use strict;
use Getopt::Std;
use Getopt::Long;
use Carp;

my %options = ();
my ($command,$src_file,$ucf_file,$src_type,$dest_file,$obj,$i_obj,$d_obj,$data_src_file,$text_src_file);

GetOptions( "init=s" => \$command,
			"in=s" => \$src_file,
			"ucf=s" => \$ucf_file,
			"in-type=s" => \$src_type,			
			"out=s" => \$dest_file,
			"obj=s" => \$obj,
			"iobj=s" => \$i_obj,
			"dobj=s" => \$d_obj,
			"text=s" => \$text_src_file,
			"data=s" => \$data_src_file);

 # if not set, set output to "-".  opening "-" for writing
 # means STDOUT
 $dest_file = "-" unless defined $dest_file;
 
 # if not set, set input to "-".  opening "-" for reading
 # means STDIN
 $src_file = "-" unless defined $src_file;

 # default object to U1
 $obj = "U1" unless defined $obj;
 
 # default command to ASM
 $command = "ASM" unless defined $command;
# print "$command $src_file $dest_file\n";
 COMMAND: {
 	$command =~ /ASM/ && do { &asm_to_hex($src_file,$dest_file); last COMMAND; };
	$command =~ /ROM\b/ && do { open(ROM,">$dest_file") or die ("Can't open dest file $dest_file\n");
								my $instructions = &assemble($src_file);
							  	my $rom_output = &rom_init($obj,$instructions,'');
								print ROM join("",@$rom_output);
								close(ROM);
							  	last COMMAND;
							 };
	$command =~ /ROM_SDC/ && do { open(ROM,">$dest_file") or die ("Can't open dest file $dest_file\n"); 
								  my $instructions = &assemble($src_file);
							  	  my $rom_output = &rom_init($obj,$instructions,'SDC');
								  print ROM join("",@$rom_output);
								  close(ROM);
							  	  last COMMAND;
							 };
    $command =~ /UCF\b/ && do { 
								&update_ucf_file($src_file,$dest_file,$ucf_file,$obj);
							  	last COMMAND;
							 };							 

	$command =~ /BYTE/ && do { &byte_ram_init($src_file,$dest_file,$obj); last COMMAND; }; 
	$command =~ /ASYNC\b/ && do { 	open(RAM,">$dest_file") or die ("Can't open dest file $dest_file\n");
									my $ram_output = &async_ram_init($src_file,$obj,'');
									print RAM join("",@$ram_output);
									close(RAM);
									last COMMAND; }; 
	$command =~ /ASYNC_SDC/ && do { open(RAM,">$dest_file") or die ("Can't open dest file $dest_file\n");
									my $ram_output = &async_ram_init($src_file,$obj,'SDC');
									print RAM join("",@$ram_output);
									close(RAM);
									last COMMAND; };
	$command =~ /RAM/ && do { &ram_init($src_file,$dest_file,$obj); last COMMAND; };
	$command =~ /SPIM/ && do { open(ASM,"<$src_file") or die("can't open $src_file\n");
							   open(ROM,">$dest_file") or die ("Can't open dest file $dest_file\n"); 
							   my @lines = <ASM>;  
#							   print "lines[5] =",$lines[5],"\n";
							   my @fields = map { [ split('\s+',$_) ] } @lines;  
#   							   print "fields[5] =",join(",",@{$fields[5]}),"\n";
							   my @instructions = map { $_->[1] =~ s/^0x//; [ hex($_->[1]) ] } @fields; 
#							   print "instructions[5] =",join(",",@{$instructions[5]}),"\n";
							   my $rom_output = &rom_init($obj,\@instructions);	
							   print ROM join("",@$rom_output);
							   close(ROM);
							  }; 
	$command =~ /^SDC/ && do { open(SDC,">$dest_file") or die ("Can't open dest file $dest_file\n"); 
								  my $instructions = &assemble($text_src_file);
							  	  my $rom_output = &rom_init($i_obj,$instructions,'SDC');
								  my $ram_output = &async_ram_init($data_src_file,$d_obj,'SDC');
								  print SDC join("",@$rom_output);
								  print SDC join("",@$ram_output);
								  close(SDC);
							  	  last COMMAND;
							 };
	}
 
 exit;

sub asm_to_hex{
	my $asm_file = shift or die("Need an input asm file\n");
	my $hex_file = shift or die("Need an output file\n");
	open(HEX,">$hex_file") or die("can't open $hex_file\n");

	my $instructions = &assemble($asm_file); 
#	print join("\n",@$instructions);
	my $i;
	foreach	$i (@$instructions){
		my ($result,$instr,$opcode,$rs,$rt,$immed) = @{$i};
		$opcode = (($result & 0xFC000000)>>26);
		$rs = (($result & 0x03E00000)>>21);
		$rt = (($result & 0x001F0000)>>16);
		$immed = ($result & 0x0000FFFF);
		printf HEX ("%8.8x // %s || %d | %d | %d | %d | %d | %d ||\n",$result,$instr,$opcode,$rs,$rt,(($immed & 0x0000F800) >> 11),(($immed & 0x000007C0)>> 6),($immed & 0x0000003F));
	}
	close(HEX);
}			 

#
# RS: Simple function that takes a register specifier and returns a 32-bit word containing the correct
#	     register in the RS field (bits [25:21])
# args:  name -> one of the accepted register specifiers
# returns: register number, masked to 5 bits, inserted as bits 25:21 in an int

sub RS{
	my ($package, $filename, $line) = caller;
	my $name = shift or die("$package:$filename:$line: RS needs register input\n");
	my $num = &register_replace($name,21);
	return $num;
}

#
# RT: Simple function that takes a register specifier and returns a 32-bit word containing the correct
#	     register in the RT field (bits [20:16])
# args:  name -> one of the accepted register specifiers
# returns: register number, masked to 5 bits, inserted as bits 20:16 in an int

sub RT{
	my $name = shift or carp("make_rt needs register input\n");
	my $num = &register_replace($name,16);
	return $num;
}

#
# RD: Simple function that takes a register specifier and returns a 32-bit word containing the correct
#	     register in the RD field (bits [15:11])
# args:  name -> one of the accepted register specifiers
# returns: register number, masked to 5 bits, inserted as bits 15:11 in an int

sub RD{
	my $name = shift or carp("make_rt needs register input\n");
	my $num = &register_replace($name,11);
	return $num;
}

#
# Shift: Simple function that takes a register specifier and returns a 32-bit word containing the correct
#	     register in the ShAmt field (bits [10:6])
# args:  decimal or hexidecimal
# returns: shift amount, masked to 5 bits, inserted as bits 10:6 in an int

sub SHIFT{
	my $name = shift or carp("make_shift needs input\n");
	my $num = (&parse_number($name) << 6);
	die("Invalid argument for shift\n") unless defined($num);
	return $num;
}

sub new_sym_info {
	my $index = shift or undef;
	my $type = shift or 'WORD';
	my $uses = shift or [];
	return { type => $type, index => $index, uses => $uses };	
}

#  Convert a file containing assembly 1 to a line into the decimal values for those instructions
#  Doesn't support Labels or much fancy stuff
#  returns reference to an array containing the decimal values

sub assemble {
  	my $asm_file = shift;
	open(ASM,"<$asm_file") or die("can't open $asm_file\n");

	my @instructions = ();

	#### Create a symbol table
	#### Symbol format:
	#    Symbol => { TYPE => ( WORD | BYTE ), index => number, uses => USES }
    #    USES :=  [ USES NUM |  ]
	#
	#    Outlines: Array representing output memory format
	#              composed of INST elements
	#    INST  := { TYPE => OPTYPE, SYMBOL => SYMBOLTYPE , VALUE => Encoded Operation, LINE => line number, RAW => asm line }
	#    OPTYPE := J_OP | B_OP | IM_2_OP | IM_1_OP | A_3_OP | A_1_T_OP | A_1_F_OP | A_0_OP | SH_OP | COMMENT
	#    SYMBOLTYPE:= NONE | { 'name' => SYM_NAME, 'usage' => SYM_USE }
	#    SYM_NAME := [_]*\w[\w\d_]*
	#    SYM_USE := JMP | BR | MEM 
	#   
	my %symbols = (); 
	my $line = 0;
	my @outlines = ();
	my %inst = ();
	while(<ASM>){
		chomp;
		s/^\s+//g; #kill leading spaces
		if( m/^\s*(\/\/|\#)/s ){	   # pass comment lines into the Hex file
#			print HEX $_,"\n";
			@inst{'TYPE','LINE','RAW'} = ('COMMENT',$line,$_."\n");			
			push @outlines,{ %inst };
#			$line = $line + 1;
			next;
		} elsif( $_ eq "" ){  #skip whitespace
			next;
		} 
		my $raw_instruction = $_;	
		my @fields = split(/[\s,]+/); # tokenize the current instruction
#print "Fields = ",join(",",@fields),"\n";
eval {		my ($ID) = ($fields[0] =~ m/([\w\d]+)\s*:/);
		# check for a label
		if(defined($ID)){
			my $label = "\L$ID"; # get lower-case version of $ID
#			print "LABEL: $label at line $line\n";
			shift @fields; # remove the label from list of fields
			# This label may have already been used. I.E. as a forward declaration.
			# So we need to check the labels hash, and update any instructions that have used it
			my $sym_info = ((exists($symbols{$label}) && defined($symbols{$label}))? 
							$symbols{$label} : &new_sym_info($line,'WORD'));
			$sym_info->{index} = $line;
			$symbols{$label} = { %$sym_info };  # store this label info for later uses
			my $sym_type = $sym_info->{type}; 
#			print $label."{type} = $sym_type\n";
			my $sym_index = $sym_info->{index};
			my $uses = $sym_info->{'uses'};

			my $use;
			# replace usages of this label with the index
			foreach $use (@$uses){
				my $debug = "replacing $ID in line $use with ";	 
#				print $debug,"\n";
				$outlines[$use] = &replace_symbol($outlines[$use],$symbols{$label},$debug);
			}
			next unless ($#fields > 0);
		}
		if($fields[0] eq ""){
			shift @fields; # discard useless empty field
		}
		
#		print "Fields = ",join(",",@fields),"\n";
		# decode the word
		my ($type,$code) = &op_decode($fields[0]);
		@inst{'TYPE','LINE','RAW'} = ($type,$line,$raw_instruction);
		my $opcode = shift @fields; # move opcode from fields[0] into $opcode
#		print "opcode = $opcode\ntype = $type\n";
#		my $new_use = { 'TYPE' => undef, 'VALUE' => $code, 'LINE' => $line };
		my $result = { %inst } ; #$new_use;
#		print "new_use = $new_use\n";
		if($type =~ m/^J_OP/){
#			$new_use->{'TYPE'} = 'JMP';
			$result = { %inst };
#			print "Fields = ",join(",",@fields),"\n";
			my ($target,$sym) = &imm_or_sym($fields[0],-1); # don't mask address within imm_or_sym
			if(defined($target)){
				@{$result}{'SYMBOL','VALUE'} = ('NONE', ($code | (($target << 2) & 0x03FFFFFF)));
			} elsif(defined($sym)){
				@{$result}{'SYMBOL','VALUE'} = ({'name' => $sym, 'usage' => 'JMP'},$code);
				if(defined($symbols{$sym}) && (defined($symbols{$sym}->{index}))){
				 	  my $debug = "replaced symbol $sym with ";
 			          $result = &replace_symbol($result,$symbols{$sym},$debug);
				} else { 
#					print "added $sym to list \n";
					add_use(\%symbols,$sym,(scalar @outlines)); # update list of uses of a symbol with this line
					my ($k,$v,$k2,$v2);
					while(($k,$v) = each(%symbols)){
#						print "after add: $k -> ";
						while(($k2,$v2) = each(%{$v})){
#							print "{ $k2 => ";
							if(defined($v2)){
								if(ref($v2) eq "ARRAY"){
#									print "(",join(",",@$v2),")";
								} else {
#									print $v2;
								}
							}
#							print " }\n"; 	
						}	
					}	
				}
			} else {
				die ("Malformed instruction: $raw_instruction\n");
			}				
		} elsif( $type =~ m/^B_OP/){

			$result->{'VALUE'} = ($code | &RS($fields[0]) | RT($fields[1]));
#			print "$opcode => ",join(",",@fields),"\n";
			my ($offset,$sym) = &imm_or_sym($fields[2],0xFFFF);  # mask offset to 16 bits  
#			printf "offset = %s, sym = %s\n",(defined($offset)? "$offset" : 'undef'),(defined($sym)? "$sym" : 'undef');
			if(defined($offset)){
				@{$result}{'SYMBOL','VALUE'} = ('NONE', ( $result->{VALUE} | $offset ));
			} elsif(defined($sym)){
				$result->{'SYMBOL'} = {'name' => $sym, 'usage' => 'BR'};
			    if((defined($symbols{$sym})) &&
				 (defined($symbols{$sym}->{index}))){
				 	  my $debug = "replaced symbol $sym with ";
 			          $result = &replace_symbol($result,$symbols{$sym},$debug);
			    } else {
				 add_use(\%symbols,$sym,(scalar @outlines)); # update list of uses of a symbol with this line
			    }
			} else {
				die ("Malformed instruction: $raw_instruction\n");
			}
		 } elsif( $type =~ m/^BZ_OP/){

			$result->{'VALUE'} = ($code | &RS($fields[0]));
#			print "$opcode => ",join(",",@fields),"\n";
			my ($offset,$sym) = &imm_or_sym($fields[1],0xFFFF);  # mask offset to 16 bits  
#			printf "offset = %s, sym = %s\n",(defined($offset)? "$offset" : 'undef'),(defined($sym)? "$sym" : 'undef');
			if(defined($offset)){
				@{$result}{'SYMBOL','VALUE'} = ('NONE', ( $result->{VALUE} | $offset ));
			} elsif(defined($sym)){
				$result->{'SYMBOL'} = {'name' => $sym, 'usage' => 'BR'};
			    if((defined($symbols{$sym})) &&
				 (defined($symbols{$sym}->{index}))){
				 	  my $debug = "replaced symbol $sym with ";
 			          $result = &replace_symbol($result,$symbols{$sym},$debug);
			    } else {
				 add_use(\%symbols,$sym,(scalar @outlines)); # update list of uses of a symbol with this line
			    }
			} else {
				die ("Malformed instruction: $raw_instruction\n");
			}
		} elsif( $type =~ m/^M_OP/){
			my ($arg,$reg) = ($fields[1] =~ /([\w\d]+)\((\$[^\)]+)\)/);
			my $rs = &RS($reg);
			my $rt = &RT($fields[0]);
#			print "rs ($reg) = ".($rs >> 21).", rt ($fields[0]) = ".($rt >> 16)."\n";
			$result->{'VALUE'} = ($code | &RT($fields[0]) | &RS($reg));
#			print "mem op opcode = $opcode\tfields[0] = $fields[0]\tfields[1] = $fields[1]\targ = $arg\treg = $reg\n";
#			printf "%8.8x\n",$result->{'VALUE'};
			my ($offset,$sym) = &imm_or_sym($arg,0xFFFF);
			if(defined($offset)){
				@{$result}{'SYMBOL','VALUE'} = ('NONE', ( $result->{VALUE} | $offset ));
			} elsif(defined($sym)){
				$result->{'SYMBOL'} = {'name' => $sym, 'usage' => 'MEM'};
			    if((exists($symbols{$sym})) &&
				 (defined($symbols{$sym}->{index}))){
				 	  my $debug = "replaced symbol $sym with ";
 			          $result = &replace_symbol($result,$symbols{$sym},$debug);
			    } else {
				 	add_use(\%symbols,$sym,(scalar @outlines)); # update list of uses of a symbol with this line
			    } 
			} else {
				die ("Malformed instruction: $raw_instruction\n");
			}

			#$_ = $offset;
			#SW_MEM_OFFSET: {
			#	/^(-[0-9]+|[0-9]+)/ && do { $offset = int($1); last SW_MEM_OFFSET; };
			#	/^(0x([\dA-Fa-f]+)/ && do { $offset = hex($1); last SW_MEM_OFFSET; };
			#	/^(\w[\w\d_]*):/ && do { my $sym = $1;

		} elsif( $type =~ m/A_3_OP/){
#			print "$opcode => ",join(",",@fields),"\n";
			@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | &RD($fields[0]) | &RS($fields[1]) | RT($fields[2])));
		} elsif( $type =~ m/A_2_OP/){
			print "$opcode => ",join(",",@fields),"\n";
			@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | &RS($fields[0]) | &RT($fields[1])));
#		} elsif( $type =~ m/A_2_T_OP/){
#			print "$opcode => ",join(",",@fields),"\n";
#			@{$result}{'TYPE','VALUE'} = ('DONE',($code | &RD($fields[0]))));
		} elsif( $type =~ m/A_1_F_OP/){
#			print "$opcode => ",join(",",@fields),"\n";
			@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | &RD($fields[0])));
		} elsif( $type =~ m/A_1_T_OP/){
#			print "$opcode => ",join(",",@fields),"\n";
			@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | &RS($fields[0])));

###added by Adrian
		} elsif( $type =~ m/A_1_OP/){
			# jr, jalr, t????
			@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | &RS($fields[0])));
			
###end			
		} elsif( $type =~ m/A_0_OP/){
			# should only happen for syscall, break, nop, halt 
			@{$result}{'SYMBOL','VALUE'} = ('NONE',$code);
		} elsif($type =~ m/SH_OP/){
			@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | &RD($fields[0]) | &RT($fields[1]) | &SHIFT($fields[2])));
#			printf "shift = %8.8x\n",$result->{'VALUE'};
			#die("No support for shift operations yet!!\n");
		} elsif($type =~ m/SH_V_OP/){
			@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | &RD($fields[0]) | &RT($fields[1]) | &RS($fields[2])));
#			printf "shift = %8.8x\n",$result->{'VALUE'};
		} elsif($type =~ m/IM_2_OP/){
#			print "$opcode => ",join(",",@fields),"\n";
			my ($imm,$sym) = &imm_or_sym($fields[2],0xFFFF);
			if(defined($imm)){
				@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | RS($fields[1]) | RT($fields[0]) | $imm)); 
			} else {
			   die("Don't support symbols as immediates yet");
			}
		} elsif($type =~ m/IM_1_OP/){ # LUI $t0, 12
			my ($imm,$sym) = &imm_or_sym($fields[1],0xFFFF);
			if(defined($imm)){
				@{$result}{'SYMBOL','VALUE'} = ('NONE',($code | 0 | &RT($fields[0]) | $imm)); 
			} else {
			   	die("Don't support symbols as immediates yet");
			}
		}
#		printf "result[TYPE] = %s\tresult[VALUE] = %x\topcode = %x\trs = %x\trt = %x\timm = %x\n",
#				$result->{'TYPE'},
#				$result->{'VALUE'},
#				(($result->{'VALUE'} >> 26)&0x0000003F),
#				(($result->{'VALUE'} >> 21)&0x0000001F),
#				(($result->{'VALUE'} >> 16)&0x0000001F),
#				($result->{'VALUE'} & 0x0000FFFF); 

		push @outlines,$result;
		$line = $line + 1;
	}; carp ("Error in instruction $raw_instruction:\n$@\n") if $@;
#	my ($opcode,$rs,$rt,$immed,$result) = &op_replace(@fields);
#	push @instructions,[$result,$instr,$opcode,$rs,$rt,$immed];
	}
	close(ASM);
	my $i;
	foreach $i (@outlines){
		next if($i->{'TYPE'} eq 'COMMENT');
		my ($code,$instr) = @{$i}{'VALUE','RAW'};
		my $out_instr = [ $code, $instr,
						(($code >> 26)&0x0000003F),
						(($code >> 21)&0x0000001F),
						(($code >> 16)&0x0000001F),
						($code & 0x0000FFFF) ];  
		push @instructions, $out_instr;
	}							
	return [ @instructions ];
}	

sub replace_symbol{
	my $instr = shift;
	my $sym_info = shift;
	my $debug = shift or "";
	my $sym = $instr->{'SYMBOL'};
	my ($sym_name,$sym_usage) = @{$sym}{'name','usage'};
	$_ = $sym_usage; 
	my $value;
	SWITCH_TYPE:{
		/JMP/ && do { $value = (0x03FFFFFF & ($sym_info->{index}));
				  	last; };
		/BR/ && do { my $addr = $instr->{'LINE'};
					 my $next = $addr + 1;
					 my $sym_addr = $sym_info->{index};
					 $value = (($sym_addr - $next) & 0x0000FFFF);
# 					 print "BR at line $addr\t next = $next\t sym at $sym_addr\toffset = ".($sym_addr - $next)."\n";
				 	 last; };
		/MEM/ && do { $value = ($sym_info->{index} & 0x0000FFFF);
				  	 last; };
		die("INVALID use of symbol\n");
	}
    @{$instr}{'SYMBOL','VALUE'} = ('NONE',($instr->{'VALUE'} | $value));
#	printf("%s %x\n",$debug,$value) unless ($debug eq "");
	return $instr;
}

sub add_use {
	my $symbols_ref = shift;
	my $sym = shift;
	my $line = shift;				   
#	print "Forward decl of sym $sym at line $line\n";
	my $sym_info_ref = (exists($symbols_ref->{$sym}) && defined($symbols_ref->{$sym}))?
			$symbols_ref->{$sym} : &new_sym_info();
	my $uses_ref = $sym_info_ref->{uses};
	push @$uses_ref,$line;
	$sym_info_ref->{uses} = $uses_ref;
	$symbols_ref->{$sym} = $sym_info_ref;
}

sub dec2bin {
	my $input = shift;
	my $wide = shift or 16;
	my $fmt = shift or "n";
	my $fmt2 = "B32";
    my $str = unpack($fmt2, pack($fmt, $input));
   return substr($str,-$wide);
}

sub bin2dec { 
	my $input = shift;
    return unpack("N", pack("B32", substr("0" x 32 . $input, -32)));
}		

sub mips_pack{
#	print "called mips_pack with ",join(",",@_),"\n";
	my $opcode = dec2bin(shift,6,"n");
	my $result = $opcode;
	if($opcode >= 2 && $opcode <= 3){
		$result .= &dec2bin(shift,26,"N");
	} else {
		my ($rs,$rt) = map { dec2bin($_,5,"n"); }  @_[0,1];
		if($opcode == 0){
			my ($rd,$shift) = map { dec2bin($_,5,"n"); } @_[2,3];
			my $func = dec2bin($_[4],6,"n");
#			print "R-type: ",join("|","",$opcode,$rs,$rt,$rd,$shift,$func,"\n");
			$result = join("",$opcode,$rs,$rt,$rd,$shift,$func);
		} else {
			my $immed = dec2bin($_[2],16,"n");
			$result .= $rs.$rt.$immed;
#			print "I-type: ",join("|","",$opcode,$rs,$rt,$immed,"\n");
		}
	}
	return $result;
}

#
# given a file that satisfies the following grammar:
#	FILE := LINES
# 	LINES := LINE LINES | LINE
# 	LINE  := ADDR | VALUE
# 	ADDR  := @VALUE
# 	VALUE := NUM | HEX
#
#   Convert into a sparse memory structure that has values only for specified addresses
#

sub parse_data_input {
		my $hex_file = shift;
		my $wordsize = shift or 4;
		open(HEX,"<$hex_file") or die ("Can't open memory init file $hex_file");
		my %mem = ();
		my $addresses = 2048;
		my $addr = 0;
	while(<HEX>){
		chomp;
#		print "matched $_ as ";
		/^\s*([\/][\/]|\#).*/ && do { #print "comment\n";
								 next; };  # it's a comment. Skip this line
		s/(\/\/|\#).*\$//s; # Remove line ending comments
		/^\s*@(\d+)(?!')/g && do { #print "address $1\n";
								   $addr = $1; next; }; #decimal address
		/^\s*@\d*'h([\da-fA-Fxz])/g && do { #print "address $1\n";
											$addr = hex($1); next; }; # hex address
		/^\s*@\d*'b([10xz]+)/g && do { #print "binary address $1\n";
									   die("No support for binary addresses. Convert to hex"); };
		/^\s*@\d*'d(\d+)/g && do { #print "decimal address $1\n";
								   $addr = $1; next; }; # decimal address comand
		/^\s*(\d+'h|0x)*([0-9A-Fa-f]+)/ && do { my $val = substr(("0" x 8).$2,-8);
											 #print "hex value $val\n";
											 my $dec = hex($val);
											 my $bin = dec2bin($dec,32,"N");
											 my @bits = reverse split(//,$bin);
				#							 printf "val = %s\tdec = %d\tbin = %s\tbits = %s\n",$val,$dec,$bin,join(",",reverse @bits);
											 $mem{$addr} = [ @bits ];
											 $addr += 4; next; }; # // it's a value 
	} 							  
	return \%mem;
}

sub byte_ram_init {
	my ($hex_file,$ram_file,$component) = @_[0..2];												  											
	open(RAM,">$ram_file") or die ("Can't open init file $ram_file");
	my $mem = &parse_data_input($hex_file,1);	
	# Mem structure is ready
	# Now we convert it into INIT_* = * Directives
	my @addresses = sort { $a <=> $b } keys %$mem;	
#	print join(",",@addresses),"\n";
	my $c_addr; 
	my %inits = ();
	my @o_byte_column = "x" x 256; # 8 contiguous words in the ram
	my $o_addr = 0; # byte address of existing line 
	
#	print join(',',@addresses);
	foreach $c_addr (@addresses){
		my $c_word_address = ($c_addr >> 2);
		my $c_byte; # = ($c_addr & 0x00000003);
		foreach $c_byte (0..3){
			my $c_init = ($c_word_address >> 5);
			my $c_index = ($c_word_address & 0x0000001F);
#		print "oct_addr = $c_octword_addr\toct_index = $c_octword_index\n";
#		if($c_init[$c_ != $o_addr){ # doesn't go in current line. Save current, then get new line
		
			@o_byte_column = (defined($inits{$c_init}{$c_byte}))? @{$inits{$c_init}{$c_byte}} : ("x") x 256;
			my $bit_array = $mem->{$c_addr};
			my ($hi_bit,$lo_bit) = ((($c_byte << 3) + 7),($c_byte << 3));
#			printf "bit_array = %s , hi = %d, lo = %d \n",join("",reverse @$bit_array),$hi_bit,$lo_bit;
			my @bits = @{$bit_array}[($lo_bit..$hi_bit)]; 
#			printf "addr = %x; word_addr = %x, byte = %d; init = %x; c_index = %x; val = %s; hex = %2.2X\n",$c_addr,$c_word_address,$c_byte,$c_init,$c_index,join("",reverse @bits),bin2dec(join("",reverse @bits));
			my @old_bits = splice @o_byte_column,($c_index << 3),8,@bits; 
#			printf "replacing %s at %d..%d with %s\n",join("",@old_bits),($c_index << 3),($c_index << 3)+7,join("",reverse @bits);

#			printf "%s \n",join("",reverse @o_byte_column);
			$inits{$c_init}{$c_byte} = [ @o_byte_column ];
		}
	}

    # Words have been grouped into octwords
	# Now to output proper INIT_* commands
	my ($line,$byte);
	my $first = 1; 
	print RAM "defparam\n";
	foreach $line (sort keys %inits){
		foreach $byte (0..3){ 
			my $byte_column = $inits{$line}{$byte};
#			print "byte_column size = $#$byte_column\n";
#			print join(",",@$byte_column),"\n";
			my ($j,$k);
			my $oct_string = "";
			my @hex_array = "0" x 32;
			for($k = 0; $k < 256; $k += 8){
				my @bits = @{$byte_column}[($k..($k+7))];
	   			$hex_array[($k >> 3)] = sprintf("%2.2x",&bin2dec(join("",reverse @bits)));
			} 

#  		   	print "byte_column size = $#hex_array\n";

		   $oct_string = join('',(reverse @hex_array));
#			print "\n$oct_string\n";
#			printf RAM ("%s\t%s.Byte%d.INIT_%2.2X = 256'h%s",(($first>0)? "":",\n"),$component,($byte),$line,$oct_string);
			printf RAM ("%s\t%s.BYTE%d_INIT_%2.2X = 256'h%s",(($first>0)? "":",\n"),$component,($byte),$line,$oct_string);
			$first = 0;
		}
	}
	print RAM ";\n";
}

sub new_async_init{
	my $chips = 8;
	my $bytes = 4;
	my $inits = 8;
	my ($i,$j,$k);
	my %bytes = ();
	for($i = 0; $i < $bytes; $i = $i + 1){
		my %chips = ();
		my @null = (0) x 32;
		for($j = 0; $j < $chips; $j = $j + 1){
			my %inits = (0 => [ @null ],
						 1 => [ @null ],
						 2 => [ @null ],
						 3 => [ @null ],
						 4 => [ @null ],
						 5 => [ @null ],
						 6 => [ @null ],
						 7 => [ @null ]);
#			print "inits{0} = ",join(",",@{$inits{0}});
			$chips{$j} = { %inits };	
		}
		$bytes{$i} = { %chips };
	}
	return %bytes;
}

sub async_ram_init {
	my ($hex_file,$component,$out_mode) = @_[0..2];			
	my $mem = &parse_data_input($hex_file,1);	
	# Mem structure is ready
	# Now we convert it into INIT_* = * Directives
	my @addresses = sort { $a <=> $b } keys %$mem;	
#	print join(",",@addresses),"\n";
	my $c_addr; 
	my %inits = &new_async_init(); 
	my @ram_output;
#	my @o_byte_column = "x" x 256; # 8 contiguous words in the ram
#	my $o_addr = 0; # byte address of existing line 
	

	# Input is a set of (Addr,32bit) pairs
	# Memory = 4 banks of 256 words x 8 bits
	# 1 Bank = 8 32 x 8 bit ram
	# 1 Ram = 32 8 bit words, INIT_00 - INIT_07 each bit corresponds to a word
	#
	# Addr Breakdown
	# |				|9		 2|1	0|
	# |             |  Word   | byte |
	# Word Breakdown
	# |9  chip  7|6  Word Addr  2|
	#
	# Init Structure:
	# { byte -> { chip -> { 0 => 32 bits ( 1 / word ), 1 => ... } } };
	#
	
	foreach $c_addr (@addresses){
		my $c_word_address = (($c_addr & 0x0000007C) >> 2);
		my $c_byte; # = ($c_addr & 0x00000003);
		foreach $c_byte (0..3){
			my $c_chip = (($c_addr & 0x000001C0) >> 7);
			my $chip_init_strings = $inits{$c_byte}{$c_chip};
			my $bit_array = $mem->{$c_addr};
			my ($hi_bit,$lo_bit) = ((($c_byte << 3) + 7),($c_byte << 3));
#			printf "bit_array = %s , hi = %d, lo = %d \n",join("",reverse @$bit_array),$hi_bit,$lo_bit;
			my @bits = @{$bit_array}[($lo_bit..$hi_bit)]; 
#			printf "addr = %x; word_addr = %x, byte = %d; chip = %x; val = %s; hex = %2.2X\n",$c_addr,$c_word_address,$c_byte,$c_chip,join("",reverse @bits),bin2dec(join("",reverse @bits));
			my $i;
	

			for($i = 0; $i < 8; $i = $i + 1){
				my $array = $chip_init_strings->{$i};
#				print "array = ",join(",",@$array),"\n";
#				print "pre update: ".$c_word_address," ",join(",",reverse @$array),"\n"; 

				$array->[$c_word_address] = $bits[$i];  
#				print "post update: ".$c_word_address," ",join(",",reverse @$array),"\n";
				$inits{$c_byte}{$c_chip}{$i} = $array;
			}
			
			#			printf "replacing %s at %d..%d with %s\n",join("",@old_bits),($c_index << 3),($c_index << 3)+7,join("",reverse @bits);

#			printf "%s \n",join("",reverse @o_byte_column);
		}
	}

    # Words have been grouped into octwords
	# Now to output proper INIT_* commands
	my ($line,$byte,$chip);
	my $first = 1; 
	if($out_mode ne 'SDC'){
		push @ram_output, "defparam\n";	
	}
	foreach $byte (0..3){
		foreach $chip (0..7){
			my $chip_inits = $inits{$byte}{$chip};

			my $bit;
			#added by Adrian
			if ($out_mode eq 'SDC') { $first = 1; }
			#$first = 1;
			#end
			my $any = 0;
			foreach	$bit (0..7){
				my $bit_array = $chip_inits->{$bit};
				my $dec = &bin2dec(join("", reverse @$bit_array));

				if($dec != 0){	 # skip this init if the value is zero
					$any |= 1;
					if($out_mode eq 'SDC'){
						if($first){
							push @ram_output, sprintf ("define_attribute {%s.Byte%d.U%d} xc_props {",$component,($byte),($chip+1)); 
						}
	   					push @ram_output, sprintf ("%sINIT_%2.2X = 32'h%8.8x",(($first>0)? "":",\n\t"),$bit,$dec);
					} else {
						push @ram_output, sprintf ("%s\t%s.Byte%d.U%d.INIT_%2.2X = 32'h%8.8x",(($first>0)? "":",\n"),$component,($byte),($chip+1),$bit,$dec);
					}
					$first = 0;	
				}
				push @ram_output, "}\n" if(($out_mode eq 'SDC') && ($bit == 7) && ($any));
			}
		}
	}
	if($out_mode ne 'SDC'){
		push @ram_output, ";\n";
	}							
	return \@ram_output;
}

#sub ram_init {
#	my ($hex_file,$ram_file,$component) = @_[0..2];												  											
#	open(RAM,">$ram_file") or die ("Can't open init file $ram_file");
#	my $mem = &parse_data_input($hex_file);	
#	# Mem structure is ready
#	# Now we convert it into INIT_* = * Directives
#	my @addresses = sort keys %$mem;	
##	print join(",",@addresses),"\n";
#	my $c_addr; 
#	my %inits = ();
#	my @o_line = (0) x 8; # 8 contiguous words in the ram
#	my $o_addr = 0; # octword address of existing line
#	foreach $c_addr (@addresses){
#		my ($c_octword_addr,$c_octword_index) = (($c_addr >> 3),(7-($c_addr & 0x00000007)));
##		print "oct_addr = $c_octword_addr\toct_index = $c_octword_index\n";
#		if($c_octword_addr != $o_addr){ # doesn't go in current line. Save current, then get new line
#			$inits{$o_addr} = [ @o_line ];
#			$o_addr = $c_octword_addr;
#			@o_line = (defined($inits{$c_octword_addr}))? @{$inits{$o_addr}} : (0) x 8;
#		}
#		$o_line[$c_octword_index] = $mem->{$c_addr}; 
##		print join(",",@o_line),"\n";
#	}
#	$inits{$o_addr} = [ @o_line ];
#
##	($k,$v) = (0,0);
##	while(($k,$v) = each %inits){
##		print "$k => ",$v,"\n";
##	}
#
#    # Words have been grouped into octwords
#	# Now to output proper INIT_* commands
#	my $line;
#	my $first = 1; 
#	print RAM "defparam\n";
#	foreach $line (sort keys %inits){
#		my $octword = $inits{$line};
#		my $oct_string = join("_",map { sprintf("%8.8x",$_) } @$octword);
#		printf RAM ("%s\t%s.INIT_%2.2X = 256'h%s",(($first>0)? "":",\n"),$component,$line,$oct_string);
#		$first = 0;
#	}
#	print RAM ";\n";
#}	

sub update_ucf_file {
 	my ($src_file,$dest_file,$ucf_file,$obj) = @_;
	# get the current contents of the ucf file
	open(UCF,"<$ucf_file") or die ("Can't open the ucf file: $ucf_file\n");
	my @ucf_lines = grep(!/^\s*INST/,<UCF>);
	# now close the ucf file
	close(UCF);
	# open the destination file
	open(ROM,">$dest_file") or die ("Can't open dest file $dest_file\n");
	
	# parse the src_file
	my $instructions = &assemble($src_file); 
	
	# perform the transformation from instructions to ROM format
	my $rom_output = &rom_init($obj,$instructions,'');
	
	# write the results to the destination file and close it
	print ROM @ucf_lines,join("",@$rom_output);
	close(ROM);
}

sub rom_init {
	my $component = shift;
	my $instructions = shift; 
	my $out_mode = shift;  
	my @rom_output; 
	my @inits = [] x 32; # one init for each bit of the instruction	
	my ($inst,$i) = (undef) x 2 ;
	my @slice = (0) x 256;
	foreach $i (@inits) {
			$inits[$i] = ();
	} 
	my $count = 0;							  
	my @inst_as_bits = ();
	foreach $inst (map { @{$_}[0] } @$instructions){
		my $inst_dec = $inst;
		my $inst_bin = &dec2bin($inst_dec,32,"N");
#		printf ("Inst(hex)\t%8.8x\nInst(dec)\t%u\nInst(bin)\t%s\n",$inst,$inst_dec,$inst_bin);
		my @bits = split(//,$inst_bin);	
#		print "bits = ",join(",",@bits),"\n";
		push @inst_as_bits,\@bits;
	}
	my @word;
	my $j = 31;	
	foreach $i (0..$#inst_as_bits){
			@word = @{$inst_as_bits[$i]}; 
#			print @word;
			for($j = 31; $j >=0; $j = $j - 1) {
#			my $bit = $bits[$i]; 
#			print "pre col $i = ",join(",",@{$inits[$i]}),"\n"; 
			my @init_array = (defined($inits[$j]))? @{$inits[$j]} : (0) x 256;
#			print @init_array,"\n";
#			print "init_string_len = ",$#init_array,"\n";
#			print "init_string $i = ", join(",",@init_array),"\n";
			$init_array[$i] = @word[$j];	  
#			print "init_string $i = ", join(",",@init_string),"\n";
#			$inits[$i][$count] = $bit; 
			$inits[$j] = [ @init_array ];
#    		print "post col $i = ",join("",@{$inits[$i]}),"\n";

			}
			my @inst_bits = ();
			foreach $j(0..31){
				push @inst_bits,$inits[$j][$i];
				}
#			printf "init_as_bits = %s\n",join(',',@word);
#			printf "init thinks  = %s\n",join(',',@inst_bits);
	}
	if($out_mode ne 'SDC'){
		#print ROM "defparam\n";
		#push @rom_output,"defparam\n";
	}
	foreach $j (0..31){
 		my @init_hex = (0) x 64;
		my $col_bit_string = join("",@{$inits[$j]});
		my @byte_strings = unpack('a8' x 32,$col_bit_string); # split 256 character string into 64 8-character strings
		my $col_binary = pack('b8' x 32,reverse @byte_strings);
		my ($col_hex) = unpack('H64',$col_binary); 
#		printf "col_bit_string[%d] = %s\nbyte strings = %s\ncol_hex = %s\n",$col_bit_string,join(',',@byte_strings),$col_hex;
		@init_hex = split(//,$col_hex);
		if($out_mode eq 'SDC'){
#			printf ROM "define_attribute {".$component.".U1_ARRAY%d } xc_props { INIT = 256'h%s }\n",(31-$j),join("",@init_hex);
			push @rom_output, sprintf ("define_attribute {".$component.".U1_ARRAY%d } xc_props { INIT = \"256'h%s }\n",(31-$j),join("",@init_hex));
		} else {
#			printf ROM "\t".$component." = 256'h%s%s\n",(31-$j),join("",@init_hex),(($j == 31)? ";" : ",");
			push @rom_output, sprintf ("INST \"".$component."/U1_ARRAY%d\" INIT = \"256'h%s%s\n",(31-$j),join("",@init_hex),(($j == 31)? "\";" : "\";"));
		}
	}
#	close(ROM);
	return \@rom_output;
}	

sub op_decode {
	my $op = shift;
	my ($out,$func,$mem_op,$type,$value)=(0) x 5;
	$_ = "\L$op";
	s/\s*//g;
	SWITCH: {
		/addiu/ && do { $out = 9; $type = 'IM_2_OP'; last; };				   #001001
		/addi/ && do { $out = 8; $type = 'IM_2_OP'; last; };				   #001000
		/addu/ && do {$func = 33; $type = 'A_3_OP'; last; };
		/add/  && do {$func = 32; $type = 'A_3_OP'; last; };
		/andi/ && do { $out = 12; $type = 'IM_2_OP'; last; };				   #001100
		/and/ && do { $func = 36; $type = 'A_3_OP'; last; };
		/beq/ && do { $out = 4; $type = 'B_OP'; last; };
		/bne/ && do { $out = 5; $type = 'B_OP'; last; };
		/blez/ && do { $out = 6; $type = 'BZ_OP'; last; };
		/bgtz/ && do { $out = 7; $type = 'BZ_OP'; last; };
		/jal\b/ && do { $out = 3; $type = 'J_OP'; last; };
		/jalr/ && do { $func = 9; $type = 'A_1_OP'; last; };
		/jr/ && do { $func = 8; $type = 'A_1_OP'; last; };
		/j/ && do { $out = 2; $type = 'J_OP'; last;	};
		/lbu/ && do { $out = 36; $type = 'M_OP'; $mem_op = 1; last; };
		/lhu/ && do { $out = 37; $type = 'M_OP'; $mem_op = 1; last; };
		/lb/ && do { $out = 32; $type = 'M_OP'; $mem_op = 1; last; };
 		/lh/ && do { $out = 33; $type = 'M_OP'; $mem_op = 1; last; };
		/lwl/ && do { $out = 34; $type = 'M_OP'; $mem_op = 1; last; };
		/lwr/ && do { $out = 38; $type = 'M_OP'; $mem_op = 1; last; };
		/lw/ && do { $out = 35; $type = 'M_OP'; $mem_op = 1; last; };
		/sb/ && do { $out = 40; $type = 'M_OP'; $mem_op = 1; last; };
		/sh/ && do { $out = 41;  $type = 'M_OP'; $mem_op = 1;last; };
		/swl/ && do { $out = 42;  $type = 'M_OP'; $mem_op = 1;last; };
		/swr/ && do { $out = 46;  $type = 'M_OP'; $mem_op = 1;last; };
		/sw/ && do { $out = 43;  $type = 'M_OP'; $mem_op = 1;last; };
		/sltiu/ && do { $out = 11; $type = 'IM_2_OP'; last; };	 		# 001011
		/slti/ && do { $out = 10; $type = 'IM_2_OP'; last; };			# 001010
		/sltu/ && do { $func = 43; $type = 'A_3_OP'; last; };
		/slt/ && do { $func = 42; $type = 'A_3_OP'; last; };
		/ori/ && do { $out = 13; $type = 'IM_2_OP'; last; };			# 001101
		/xori/ && do { $out = 14; $type = 'IM_2_OP'; last; };			# 001110
		/lui/ && do { $out = 15; $type = 'IM_1_OP'; last; };			# 001111 
		/sllv/ && do { $func = 4; $type = 'SH_V_OP'; last; };
		/srlv/ && do { $func = 6; $type = 'SH_V_OP'; last; };
		/srav/ && do { $func = 7; $type = 'SH_V_OP'; last; };
		/sll/ && do { $func = 0; $type = 'SH_OP'; last; };
		/srl/ && do { $func = 2; $type = 'SH_OP'; last; };
		/sra/ && do { $func = 3; $type = 'SH_OP'; last; };
		/movz/ && do { $func = 10; $type = 'A_3_OP'; last; };
		/movn/ && do { $func = 11; $type = 'A_3_OP'; last; };
		/multu/ && do { $func = 25; $type = 'A_2_OP'; last; };
		/mult/ && do { $func = 24; $type = 'A_2_OP' ; last; };
		/syscall/ && do { $func = 12; $type = 'A_0_OP'; last; };
		/break/ && do { $func = 13; $type = 'A_0_OP'; last; };
		/sync/ && do { $func = 15; $type = 'A_0_OP'; last; };
		/mfhi/ && do { $func = 16; $type = 'A_1_F_OP'; last; };
		/mthi/ && do { $func = 17; $type = 'A_1_T_OP'; last; };
		/mflo/ && do { $func = 18; $type = 'A_1_F_OP'; last; };
		/mtlo/ && do { $func = 19; $type = 'A_1_T_OP'; last; };
		/subu/ && do { $func = 35; $type = 'A_3_OP'; last; };
		/sub/  && do { $func = 34; $type = 'A_3_OP'; last; };
		/xor/ && do { $func = 38; $type = 'A_3_OP'; last; };
		/nor/ && do { $func = 39; $type = 'A_3_OP'; last; };
		/or/ && do { $func = 37; $type = 'A_3_OP'; last; };
		/tgeu/ && do { $func = 49; $type = 'A_1_OP'; last; };
		/tge/  && do { $func = 48; $type = 'A_1_OP'; last; };
		/tltu/ && do { $func = 51; $type = 'A_1_OP'; last; };
		/tlt/ && do { $func = 50; $type = 'A_1_OP'; last; };
		/teq/ && do { $func = 52; $type = 'A_1_OP'; last; };
		/tne/ && do { $func = 54; $type = 'A_1_OP'; last; };
		/halt/ && do { $out = 63; $type = 'A_0_OP'; last; };
		/nop/ && do { $func = 0; $type = 'A_0_OP'; last; };
	}		
	$value = ((($out << 26) & 0xFC000000) | ($func & 0x0000003F));
	#printf "%s > %d, %d %8.8x\n",$_,$out,$func,$value;
 	return ($type,$value);
}

sub op_replace {
	my @fields = @_;
	my ($out,$rs,$rt,$rd,$immed,$func,$mem_op)=(0) x 7;
	$_ = "\L$fields[0]";
	s/\s*//g;
	SWITCH: {
		/addiu/ && do { $out = 9; last; };				   #001001
		/addi/ && do { $out = 8; last; };				   #001000
		/addu/ && do {$func = 33; last; };
		/add/  && do {$func = 32; last; };
		/andi/ && do { $out = 12; last; };				   #001100
		/and/ && do { $func = 36; last; };
		/beq/ && do { $out = 4; last; };
		/bne/ && do { $out = 5; last; };
		/jal/ && do { $out = 3; last; };
		/jr/ && do { $func = 8; last; };
		/j/ && do { $out = 2; last;	};
		/lbu/ && do { $out = 36; $mem_op = 1; last; };
		/lhu/ && do { $out = 37; $mem_op = 1; last; };
		/lb/ && do { $out = 32; $mem_op = 1; last; };
 		/lh/ && do { $out = 33; $mem_op = 1; last; };
		/lwl/ && do { $out = 34; $mem_op = 1; last; };
		/lwr/ && do { $out = 38; $mem_op = 1; last; };
		/lw/ && do { $out = 35; $mem_op = 1; last; };
		/sb/ && do { $out = 40; $mem_op = 1; last; };
		/sh/ && do { $out = 41;  $mem_op = 1;last; };
		/swl/ && do { $out = 42;  $mem_op = 1;last; };
		/swr/ && do { $out = 46;  $mem_op = 1;last; };
		/sw/ && do { $out = 43;  $mem_op = 1;last; };
		/sltiu/ && do { $out = 11; last; };	 		# 001011
		/slti/ && do { $out = 10; last; };			# 001010
		/sltu/ && do { $func = 43; last; };
		/slt/ && do { $func = 42; last; };
		/ori/ && do { $out = 13; last; };			# 001101
		/xori/ && do { $out = 14; last; };			# 001110
		/lui/ && do { $out = 15; last; };			# 001111
		/sll/ && do { $func = 0; last; };
		/srl/ && do { $func = 2; last; };
		/sra/ && do { $func = 3; last; };
		/sllv/ && do { $func = 4; last; };
		/srlv/ && do { $func = 6; last; };
		/srav/ && do { $func = 7; last; };
		/jalr/ && do { $func = 9; last; };
		/movz/ && do { $func = 10; last; };
		/movn/ && do { $func = 11; last; };
		/syscall/ && do { $func = 12; last; };
		/break/ && do { $func = 13; last; };
		/sync/ && do { $func = 15; last; };
		/mfhi/ && do { $func = 16; last; };
		/mthi/ && do { $func = 17; last; };
		/mflo/ && do { $func = 18; last; };
		/mtlo/ && do { $func = 19; last; };
		/subu/ && do { $func = 35; last; };
		/sub/  && do { $func = 34; last; };
		/xor/ && do { $func = 38; last; };
		/nor/ && do { $func = 39; last; };
		/or/ && do { $func = 37; last; };
		/tgeu/ && do { $func = 49; last; };
		/tge/  && do { $func = 48; last; };
		/tltu/ && do { $func = 51; last; };
		/tlt/ && do { $func = 50; last; };
		/teq/ && do { $func = 52; last; };
		/tne/ && do { $func = 54; last; };
		/halt/ && do { $out = 63; last; };
	}
	my $result;
	if($mem_op){								   
		my($offset,$reg) = (0) x 2;
		$fields[2] =~ /([^(]+)\(([^)]+)\)/ && do { $offset = $1; $reg = $2;};
		$rs = register_replace($reg);
		$rt = register_replace($fields[1]);
		$immed = &immediate($offset);
		$result = bin2dec(mips_pack($out,$rs,$rt,$immed));	
	} else {
		$rs = register_replace($fields[2]);
		if($out > 0){
			$rt = register_replace($fields[1]);
			$immed = &immediate($fields[3]);
#		printf("%d | %d | %d | %d \n",$out,$rs,$rt,$immed);
#		print join(",",@fields)," ==>",&mips_pack($out,$rs,$rt,$immed),"\n";
			$result = bin2dec(mips_pack($out,$rs,$rt,$immed));	
#		print "current:        ==>",dec2bin((($out << 26)+($rs << 21)+($rt<<16)+$immed),32,"N"),"\n";
		} else {
			$rt = register_replace($fields[3]);
			$rd = register_replace($fields[1]);
			$immed = ($rd << 11)|(0x0000003F & $func);		   
#		printf("%d | %d | %d | %d | %d | %d \n",$out,$rs,$rt,$rd,0,$func);
#		print join(",",@fields)," ==>",&mips_pack($out,$rs,$rt,$rd,0,$func),"\n";
			$result = bin2dec(mips_pack($out,$rs,$rt,$rd,0,$func));
		}
	}
#	print "current:        ==>",dec2bin((($out << 26)+($rs << 21)+($rt<<16)+$immed),32,"N"),"\n";
	return ($out,$rs,$rt,$immed,$result);
}



sub register_replace {
	 $_ = shift;
	 my $shift = shift or 0;
	 my $out;
	 SWITCH: { 
		/\s*[\$]?r([12][0-9])/ && do { $out = $1 ; last; };
		/\s*[\$]?r([3][01])/ && do { $out = $1; last; };
		/\s*[\$]?r([0]?[0-9])/ && do { $out = $1 ; last; };
		/\s*[\$]?at/ && do { $out = 1; last;};
		/\s*[\$]?v([01])/  && do { $out = 2+$1; last; };
		/\s*[\$]?a([0-3])/ && do { $out = 4+$1; last; };
		/\s*[\$]?s([0-7])/ && do { $out = 16+$1; last; };
		/\s*[\$]?sp/ && do { $out = 29; last; };
		/\s*[\$]?t([0-7])/ && do { $out = 8+$1; last; };
		/\s*[\$]?t([89])/ && do { $out = 24+$1-8; last; };
		/\s*[\$]?k([01])/ && do { $out = 26+$1; last; };
		/\s*[\$]?fp/ && do { $out = 30; last; };
		/\s*[\$]?gp/ && do { $out = 28; last; };
		/\s*[\$]?ra/ && do { $out = 31; last; };
		/\s*[\$]?(rz|zero)/ && do {$out = 0; last; };
		/\s*[\$]([12][0-9])/ && do { $out = $1 ; last; };
		/\s*[\$]([3][01])/ && do { $out = $1; last; };
		/\s*[\$]([0]?[0-9])/ && do { $out = $1 ; last; };

	}
	return ( $out << $shift );
}

sub parse_number {
	my $arg = shift;
	$_ = "\L$arg";
	my $val = undef;
	SWITCH_NUM: {
  		/^(\d+)'h([\da-f_]+)/ && do { my $tmp; ($tmp = $2) =~ s/_//;  # Verilog Hex format...remove _s
									  my $mask = ((1 << $1) - 1);
									  $val = ($mask & hex($tmp));
									last SWITCH_NUM; };
		/^(\d+)'d([\d]+)/ && do { my $mask = (( 1 << $1) - 1); # bit-limited decimal
								  $val = ($mask & $2);	
								  last SWITCH_NUM; };
		/^0x([\da-f]+)/ && do { $val = hex($1);  # c-style hexidecimal
								last SWITCH_NUM; };
		/^(-\d+|\d+)/ && do { $val = $1 ;  # decimal value
					    	  last SWITCH_NUM; };		 
	}
	return $val;
}

sub imm_or_sym {
	my $arg = shift;
	my $mask = shift;
	$_ = "\L$arg"; # make it lowercase
	my ($val,$sym) = (&parse_number($arg), undef); 
#		print "imm_or_sym: $_ \n";
	SWITCH_IMM: {
		/^([_]*[\w\d_]+)/ && do { $sym = $1; 	# potentially a symbol
            				last SWITCH_IMM; };
	}
	$val = (defined($val))? ($mask & $val) : undef;
	return ($val,$sym);
} 